@@ -8,20 +8,24 @@ and making predictions with trained models
 Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
 """
 
+import copy
 import os
 import time
 import warnings
+from collections import OrderedDict
+from typing import Dict, List, Tuple, Type, Union
 
+import atomai.losses_metrics as losses_metrics_
 import numpy as np
 import torch
 import torch.nn.functional as F
-
-import atomai.losses_metrics as losses_metrics_
-from atomai.models import dilnet, dilUnet
-from atomai.utils import (Hook, cv_thresh, find_com, gpu_usage_map, img_pad,
-                          img_resize, mock_forward, peak_refinement, datatransform,
+from atomai.nets import dilnet, dilUnet
+from atomai.utils import (Hook, average_weights, cluster_coord, cv_thresh,
+                          datatransform, find_com, gpu_usage_map, img_pad,
+                          img_resize, mock_forward, peak_refinement,
                           plot_losses, preprocess_training_data, torch_format,
                           unsqueeze_channels)
+from sklearn.model_selection import train_test_split
 
 warnings.filterwarnings("ignore", module="torch.nn.functional")
 
@@ -68,10 +72,10 @@ class trainer:
         training_cycles (int):
             Number of training 'epochs' (1 epoch == 1 batch)
         model_type (str):
-            Type of model to train ('dilUnet' or 'dilnet'). See atomai.models
-            for more details.
+            Type of model to train: 'dilUnet' or 'dilnet' (Default: 'dilUnet').
+            See atomai.nets for more details.
         seed (int):
-            Deterministic mode for model training
+            Deterministic mode for model training (Default: 1)
         batch_seed (int):
             Separate seed for generating a sequence of batches
             for training/testing. Equal to 'seed' if set to None (default)
@@ -79,23 +83,27 @@ class trainer:
             Size of training and test batches
         **use_batchnorm (bool):
             Apply batch normalization after each convolutional layer
+            (Default: True)
         **use_dropouts (bool):
             Apply dropouts in the three inner blocks in the middle of a network
+            (Default: False)
         **loss (str):
             Type of loss for model training ('ce', 'dice' or 'focal')
+            (Default: 'ce')
         **upsampling_mode (str):
-            "bilinear" or "nearest" upsampling method
+            "bilinear" or "nearest" upsampling method (Default: "bilinear")
         **nb_filters (int):
             Number of convolutional filters in the first convolutional block
             (this number doubles in the consequtive block(s),
             see definition of dilUnet and dilnet models for details)
         **with_dilation (bool):
             Use dilated convolutions in the bottleneck of dilUnet
+            (Default: True)
         **layers (list):
             List with a number of layers in each block.
             For U-Net the first 4 elements in the list
             are used to determine the number of layers
-            in each block of the encoder (incluidng bottleneck layer),
+            in each block of the encoder (including bottleneck layer),
             and the number of layers in the decoder  is chosen accordingly
             (to maintain symmetry between encoder and decoder)
         **print_loss (int):
@@ -110,7 +118,7 @@ class trainer:
         **kwargs:
             One can also pass kwargs for utils.datatransform class
             to perform the augmentation "on-the-fly" (e.g. rotation=True,
-            or gauss=[20, 60])
+            gauss=[20, 60], ...)
 
     Example:
 
@@ -128,15 +136,15 @@ class trainer:
     >>> trained_model = netr.run()
     """
     def __init__(self,
-                 images_all,
-                 labels_all,
-                 images_test_all,
-                 labels_test_all,
-                 training_cycles,
-                 model_type='dilUnet',
-                 seed=1,
-                 batch_seed=None,
-                 **kwargs):
+                 images_all: Union[np.ndarray, List[np.ndarray], Dict[int, np.ndarray]],
+                 labels_all: Union[np.ndarray, List[np.ndarray], Dict[int, np.ndarray]],
+                 images_test_all: Union[np.ndarray, List[np.ndarray], Dict[int, np.ndarray]],
+                 labels_test_all: Union[np.ndarray, List[np.ndarray], Dict[int, np.ndarray]],
+                 training_cycles: int,
+                 model_type: str = 'dilUnet',
+                 seed: int = 1,
+                 batch_seed: int = None,
+                 **kwargs: Union[int, List, str, bool]) -> None:
         if seed:
             torch.manual_seed(seed)
             if torch.cuda.is_available():
@@ -187,7 +195,7 @@ class trainer:
                 "No GPU found. The training can be EXTREMELY slow",
                 UserWarning
             )
-        loss = kwargs.get('loss', "dice")
+        loss = kwargs.get('loss', "ce")
         if loss == 'dice':
             self.criterion = losses_metrics_.dice_loss()
         elif loss == 'focal':
@@ -229,9 +237,9 @@ class trainer:
         if "with_dilation" in locals():
             self.meta_state_dict["with_dilation"] = with_dilation
 
-    def dataloader(self, batch_num, mode='train'):
+    def dataloader(self, batch_num: int, mode: str = 'train') -> Tuple[torch.Tensor]:
         """
-        Generates a batch of training/test images
+        Generates 2 batches images (training and test)
         """
         # Generate batch of training images with corresponding ground truth
         if mode == 'test':
@@ -257,7 +265,7 @@ class trainer:
             images, labels = images.cuda(), labels.cuda()
         return images, labels
 
-    def train_step(self, img, lbl):
+    def train_step(self, img: torch.Tensor, lbl: torch.Tensor) -> float:
         """
         Propagates image(s) through a network to get model's prediction
         and compares predicted value with ground truth; then performs
@@ -265,23 +273,23 @@ class trainer:
         """
         self.net.train()
         self.optimizer.zero_grad()
-        prob = self.net.forward(img)
+        prob = self.net(img)
         loss = self.criterion(prob, lbl)
         loss.backward()
         self.optimizer.step()
         return loss.item()
 
-    def test_step(self, img, lbl):
+    def test_step(self, img: torch.Tensor, lbl: torch.Tensor) -> float:
         """
         Forward path for test data with deactivated autograd engine
         """
         self.net.eval()
         with torch.no_grad():
-            prob = self.net.forward(img)
+            prob = self.net(img)
             loss = self.criterion(prob, lbl)
         return loss.item()
 
-    def run(self):
+    def run(self) -> None:
         """
         Trains a neural network for *N* epochs by passing a single pair of
         training images and labels and a single pair of test images
@@ -312,11 +320,6 @@ class trainer:
                     np.around(self.test_loss[-1], 4)),
                       'GPU memory usage: {}/{}'.format(
                           gpu_usage[0], gpu_usage[1]))
-            # Save model weights (if test loss decreased)
-            if e > 0 and self.test_loss[-1] < min(self.test_loss[: -1]):
-                torch.save(self.meta_state_dict,
-                   os.path.join(self.savedir,
-                   self.savename+'_metadict_best_test_weights.tar'))
         # Save final model weights
         torch.save(self.meta_state_dict,
                    os.path.join(self.savedir,
@@ -339,8 +342,6 @@ class predictor:
     Prediction with a trained neural network
 
     Args:
-        image_data (2D or 3D numpy array):
-            Image stack or a single image (all greyscale)
         trained_model (pytorch object):
             Trained pytorch model (skeleton+weights)
         refine (bool):
@@ -373,18 +374,17 @@ class predictor:
         >>> expdata = np.load('expdata-test.npy')
         >>> # Get prediction from a trained model
         >>> # (it also returns the input to NN in case the image was resized, etc.)
-        >>> nn_input, (nn_output, coordinates) = atomnet.predictor(expdata, trained_model).run()
+        >>> nn_input, (nn_output, coords) = atomnet.predictor(trained_model).run(expdata)
 
     """
     def __init__(self,
-                 image_data,
-                 trained_model,
-                 refine=False,
-                 resize=None,
-                 use_gpu=False,
-                 logits=True,
-                 seed=1,
-                 **kwargs):
+                 trained_model: Type[torch.nn.Module],
+                 refine: bool = False,
+                 resize: Union[Tuple, List] = None,
+                 use_gpu: bool = False,
+                 logits: bool = True,
+                 seed: int = 1,
+                 **kwargs: Union[int, float, bool]) -> None:
         if seed:
             torch.manual_seed(seed)
             np.random.seed(seed)
@@ -399,23 +399,19 @@ class predictor:
             hookF = [Hook(layer[1]) for layer in list(model._modules.items())]
             mock_forward(model)
             self.nb_classes = [hook.output.shape for hook in hookF][-1][1]
-        downsampling = kwargs.get('downsampling', None)
-        if downsampling is None:
+        self.downsampling = kwargs.get('downsampling', None)
+        if self.downsampling is None:
             hookF = [Hook(layer[1]) for layer in list(model._modules.items())]
             mock_forward(model)
             imsize = [hook.output.shape[-1] for hook in hookF]
-            downsampling = max(imsize) / min(imsize)
+            self.downsampling = max(imsize) / min(imsize)
         self.model = model
         if use_gpu:
             self.model.cuda()
         else:
             self.model.cpu()
-        if image_data.ndim == 2:
-            image_data = np.expand_dims(image_data, axis=0)
-        if resize is not None:
-            image_data = img_resize(image_data, resize)
-        image_data = img_pad(image_data, downsampling)
-        self.image_data = torch_format(image_data)
+
+        self.resize = resize
         self.logits = logits
         self.refine = refine
         self.d = kwargs.get("d", None)
@@ -423,7 +419,19 @@ class predictor:
         self.use_gpu = use_gpu
         self.verbose = kwargs.get("verbose", True)
 
-    def predict(self, images):
+    def preprocess(self, image_data: np.ndarray) -> torch.Tensor:
+        """
+        Prepares an input for a neural network
+        """
+        if image_data.ndim == 2:
+            image_data = np.expand_dims(image_data, axis=0)
+        if self.resize is not None:
+            image_data = img_resize(image_data, self.resize)
+        image_data = img_pad(image_data, self.downsampling)
+        image_data = torch_format(image_data)
+        return image_data
+
+    def predict(self, images: torch.Tensor) -> np.ndarray:
         """
         Returns 'probability' of each pixel
         in image(s) belonging to an atom/defect
@@ -432,7 +440,7 @@ class predictor:
             images = images.cuda()
         self.model.eval()
         with torch.no_grad():
-            prob = self.model.forward(images)
+            prob = self.model(images)
         if self.logits:
             if self.nb_classes > 1:
                 prob = F.softmax(prob, dim=1)
@@ -450,41 +458,63 @@ class predictor:
         prob = prob.numpy()
         return prob
 
-    def decode(self):
+    def decode(self,
+               image_data: np.ndarray,
+               **kwargs: int) -> Tuple[np.ndarray]:
         """
         Make prediction
-        """
-        if self.image_data.shape[0] < 20 and min(self.image_data.shape[2:4]) < 512:
-            decoded_imgs = self.predict(self.image_data)
-        else:
-            n, _, w, h = self.image_data.shape
-            decoded_imgs = np.zeros((n, w, h, self.nb_classes))
-            for i in range(n):
-                decoded_imgs[i, :, :, :] = self.predict(self.image_data[i:i+1])
-        images_numpy = self.image_data.permute(0, 2, 3, 1).numpy()
-        return images_numpy, decoded_imgs
 
-    def run(self):
-        warn_msg = (f"The output of predictor.run() has changed from "
-        f"'nn_input, nn_output' to 'nn_input, (nn_output, coordinates)'. "
-        f"Example: 'nn_input, (nn_output, coordinates) = predictor(args).run()'")
+        Args:
+            image_data (2D or 3D numpy array):
+                Image stack or a single image (all greyscale)
+            **num_batches: number of batches
+        """
+        image_data = self.preprocess(image_data)
+        n, _, w, h = image_data.shape
+        num_batches = kwargs.get("num_batches")
+        if num_batches is None:
+            if w >= 256 or h >= 256:
+                num_batches = len(image_data)
+            else:
+                num_batches = 10
+        batch_size = len(image_data) // num_batches
+        if batch_size < 1:
+            num_batches = batch_size = 1
+        decoded_imgs = np.zeros((n, w, h, self.nb_classes))
+        for i in range(num_batches):
+            if self.verbose:
+                print("\rBatch {}/{}".format(i+1, num_batches), end="")
+            images_i = image_data[i*batch_size:(i+1)*batch_size]
+            decoded_i = self.predict(images_i)
+            decoded_imgs[i*batch_size:(i+1)*batch_size] = decoded_i
+        images_i = image_data[(i+1)*batch_size:]
+        if len(images_i) > 0:
+            decoded_i = self.predict(images_i)
+            decoded_imgs[(i+1)*batch_size:] = decoded_i
+        images_data = image_data.permute(0, 2, 3, 1).numpy()
+        return images_data, decoded_imgs
+
+    def run(self,
+            image_data: np.ndarray,
+            **kwargs: int) -> Tuple[np.ndarray, Tuple[np.ndarray]]:
+        """
+        Make prediction with a trained model and calculate coordinates
+
+        Args:
+            image_data (2D or 3D numpy array):
+                Image stack or a single image (all greyscale)
+            **num_batches: number of batches (Default: 10)
+        """
         start_time = time.time()
-        images, decoded_imgs = self.decode()
-        warnings.warn(warn_msg, UserWarning)
-        coordinates = locator(decoded_imgs, self.thresh).run()
+        images, decoded_imgs = self.decode(image_data, **kwargs)
+        loc = locator(self.thresh, refine=self.refine, d=self.d)
+        coordinates = loc.run(decoded_imgs, images)
         if self.verbose:
             n_images_str = " image was " if decoded_imgs.shape[0] == 1 else " images were "
-            print(str(decoded_imgs.shape[0])
+            print("\n" + str(decoded_imgs.shape[0])
                   + n_images_str + "decoded in approximately "
                   + str(np.around(time.time() - start_time, decimals=4))
                   + ' seconds')
-        if self.refine:
-            print('\rRefining atomic positions... ', end="")
-            coordinates_r = {}
-            for i, (img, coord) in enumerate(zip(images, coordinates.values())):
-                coordinates_r[i] = peak_refinement(img[...,0], coord, self.d)
-            print("Done")
-            return images, (decoded_imgs, coordinates_r)
         return images, (decoded_imgs, coordinates)
 
 
@@ -505,39 +535,53 @@ class locator:
     Example:
 
         >>> # Transform output of atomnet.predictor to atomic classes and coordinates
-        >>> coordinates = atomnet.locator(nn_output).run()
+        >>> coordinates = atomnet.locator(dist_edge=10, refine=False).run(nn_output)
     """
     def __init__(self,
-                 nn_output,
-                 threshold=0.5,
-                 dist_edge=5,
-                 dim_order='channel_last'):
+                 threshold: float = 0.5,
+                 dist_edge: int = 5,
+                 dim_order: str = 'channel_last',
+                 **kwargs: Union[bool, float]) -> None:
 
-        if nn_output.shape[-1] == 1: # Add background class for 1-channel data
+        self.dim_order = dim_order
+        self.threshold = threshold
+        self.dist_edge = dist_edge
+        self.refine = kwargs.get("refine")
+        self.d = kwargs.get("d")
+
+    def preprocess(self, nn_output: np.ndarray) -> np.ndarray:
+        """
+        Prepares data for coordinates extraction
+        """
+        if nn_output.shape[-1] == 1:  # Add background class for 1-channel data
             nn_output_b = 1 - nn_output
             nn_output = np.concatenate(
                 (nn_output[:, :, :, None], nn_output_b[:, :, :, None]), axis=3)
-        if dim_order == 'channel_first':  # make channel dim the last dim
+        if self.dim_order == 'channel_first':  # make channel dim the last dim
             nn_output = np.transpose(nn_output, (0, 2, 3, 1))
-        elif dim_order == 'channel_last':
+        elif self.dim_order == 'channel_last':
             pass
         else:
             raise NotImplementedError(
                 'For dim_order, use "channel_first"',
                 'or "channel_last" (e.g. tensorflow)')
-        self.nn_output = nn_output
-        self.threshold = threshold
-        self.dist_edge = dist_edge
+        return nn_output
 
-    def run(self):
+    def run(self, nn_output: np.ndarray, *args: np.ndarray) -> Dict[int, np.ndarray]:
         """
         Extract all atomic coordinates in image
         via CoM method & store data as a dictionary
         (key: frame number)
-        """
 
+        Args:
+            nn_output (4D numpy array):
+                Output (prediction) of a neural network
+            *args: 4D input into a neural network (experimental data)
+
+        """
+        nn_output = self.preprocess(nn_output)
         d_coord = {}
-        for i, decoded_img in enumerate(self.nn_output):
+        for i, decoded_img in enumerate(nn_output):
             coordinates = np.empty((0, 2))
             category = np.empty((0, 1))
             # we assume that class 'background' is always the last one
@@ -545,25 +589,35 @@ class locator:
                 decoded_img_c = cv_thresh(
                     decoded_img[:, :, ch], self.threshold)
                 coord = find_com(decoded_img_c)
-                coord_ch = self.rem_edge_coord(coord)
+                coord_ch = self.rem_edge_coord(coord, *nn_output.shape[1:3])
                 category_ch = np.zeros((coord_ch.shape[0], 1)) + ch
                 coordinates = np.append(coordinates, coord_ch, axis=0)
                 category = np.append(category, category_ch, axis=0)
             d_coord[i] = np.concatenate((coordinates, category), axis=1)
+        if self.refine:
+            if len(args) > 0:
+                imgdata = args[0]
+            else:
+                raise AssertionError("Pass input image(s) for coordinates refinement")
+            print('\n\rRefining atomic positions... ', end="")
+            d_coord_r = {}
+            for i, (img, coord) in enumerate(zip(imgdata, d_coord.values())):
+                d_coord_r[i] = peak_refinement(img[..., 0], coord, self.d)
+            print("Done")
+            return d_coord_r
         return d_coord
 
-    def rem_edge_coord(self, coordinates):
+    def rem_edge_coord(self, coordinates: np.ndarray, w: int, h: int) -> np.ndarray:
         """
         Removes coordinates at the image edges
         """
 
         def coord_edges(coordinates, w, h):
-            return [coordinates[0] > w - self.dist_edge,
+            return [coordinates[0] > h - self.dist_edge,
                     coordinates[0] < self.dist_edge,
-                    coordinates[1] > h - self.dist_edge,
+                    coordinates[1] > w - self.dist_edge,
                     coordinates[1] < self.dist_edge]
 
-        w, h = self.nn_output.shape[1:3]
         coord_to_rem = [
                         idx for idx, c in enumerate(coordinates)
                         if any(coord_edges(c, w, h))
@@ -571,3 +625,312 @@ class locator:
         coord_to_rem = np.array(coord_to_rem, dtype=int)
         coordinates = np.delete(coordinates, coord_to_rem, axis=0)
         return coordinates
+
+
+class ensemble_trainer:
+    """
+    Trains multiple deep learning models, each with its own unique trajectory
+
+    Args:
+        X_train (numpy array): Training images
+        y_train (numpy array): Training labels (aka ground truth aka masks)
+        X_test (numpy array): Test images
+        y_test (numpy array): Test labels
+        n_models (int): number of models in ensemble
+        model_type (str): 'dilUnet' or 'dilnet'. See atomai.models for details
+        training_cycles_base (int): Number of training iterations for baseline model
+        training_cycles_ensemble (int): Number of training iterations for every ensemble model
+        filename (str): Filepath for saving weights
+        **kwargs:
+            One can also pass kwargs to atomai.atomnet.trainer class for adjusting
+            network parameters (e.g. batchnorm=True, nb_filters=25, etc.)
+            and to atomai.utils.datatransform class to perform the augmentation
+            "on-the-fly" (e.g. rotation=True, gauss=[20, 60], etc.)
+    """
+    def __init__(self, X_train: np.ndarray, y_train: np.ndarray,
+                 X_test: np.ndarray = None, y_test: np.ndarray = None,
+                 n_models=30, model_type: str = "dilUnet",
+                 training_cycles_base: int = 1000,
+                 training_cycles_ensemble: int = 50,
+                 filename: str = "./model", **kwargs: Dict) -> None:
+
+        if X_test is None or y_test is None:
+            X_train, X_test, y_train, y_test = train_test_split(
+                X_train, y_train, test_size=kwargs.get("test_size", 0.15),
+                shuffle=True, random_state=0)
+        self.X_train, self.y_train = X_train, y_train
+        self.X_test, self.y_test = X_test, y_test
+        self.model_type, self.n_models = model_type, n_models
+        self.iter_base = training_cycles_base
+        self.iter_ensemble = training_cycles_ensemble
+        self.filename, self.kdict = filename, kwargs
+        self.ensemble_state_dict = {}
+
+    def train_baseline(self) -> Type[torch.nn.Module]:
+        """
+        Trains a single baseline model
+        """
+        print('Training baseline model:')
+        trainer_base = trainer(
+            self.X_train, self.y_train,
+            self.X_test, self.y_test,
+            self.iter_base, self.model_type,
+            plot_training_history=True,
+            savename=self.filename + "_base.pt",
+            **self.kdict)
+        trained_basemodel = trainer_base.run()
+
+        return trained_basemodel
+
+    def train_ensemble(self,
+                       basemodel: Union[OrderedDict, Type[torch.nn.Module]],
+                       **kwargs: Dict
+                       ) -> Tuple[Dict[int, Dict[str, torch.Tensor]], Type[torch.nn.Module]]:
+        """
+        Trains ensemble of models starting each time from baseline weights
+
+        Args:
+            basemodel (pytorch object): Baseline model or baseline weights
+            **kwargs: Updates kwargs from the ensemble class initialization
+                (can be useful for iterative training)
+        """
+        if len(kwargs) != 0:
+            for k, v in kwargs.items():
+                self.kdict[k] = v
+        if isinstance(basemodel, OrderedDict):
+            initial_model_state_dict = copy.deepcopy(basemodel)
+        else:
+            initial_model_state_dict = copy.deepcopy(basemodel.state_dict())
+        n_models = kwargs.get("n_models")
+        if n_models is not None:
+            self.n_models = n_models
+        filename = kwargs.get("filename")
+        training_cycles_ensemble = kwargs.get("training_cycles_ensemble")
+        if training_cycles_ensemble is not None:
+            self.iter_ensemble = training_cycles_ensemble
+        if filename is not None:
+            self.filename = filename
+        print('Training ensemble models:')
+        for i in range(self.n_models):
+            print('Ensemble model', i+1)
+            trainer_i = trainer(
+                self.X_train, self.y_train, self.X_test, self.y_test,
+                self.iter_ensemble, self.model_type, batch_seed=i,
+                print_loss=10, plot_training_history=False, **self.kdict)
+            self.update_weights(trainer_i.net.state_dict().values(),
+                                initial_model_state_dict.values())
+            trained_model_i = trainer_i.run()
+            self.ensemble_state_dict[i] = trained_model_i.state_dict()
+
+        ensemble_metadict = copy.deepcopy(trainer_i.meta_state_dict)
+        ensemble_metadict["weights"] = self.ensemble_state_dict
+        torch.save(ensemble_metadict, self.filename + "_ensemble.tar")
+
+        ensemble_state_dict_aver = average_weights(self.ensemble_state_dict)
+        ensemble_aver_metadict = copy.deepcopy(trainer_i.meta_state_dict)
+        ensemble_aver_metadict["weights"] = ensemble_state_dict_aver
+        torch.save(ensemble_aver_metadict, self.filename + "_ensemble_aver_weights.pt")
+
+        trainer_i.net.load_state_dict(ensemble_state_dict_aver)
+
+        return self.ensemble_state_dict, trainer_i.net
+
+    @classmethod
+    def update_weights(cls,
+                       statedict1: Dict[str, torch.Tensor],
+                       statedict2: Dict[str, torch.Tensor]) -> None:
+        """
+        Updates (in place) state dictionary of pytorch model
+        with weights from another model with the same structure;
+        skips layers that have different dimensions
+        (e.g. if one model is for single class classification
+        and the other one is for multiclass classification,
+        then the last layer wights are not updated)
+        """
+        for p1, p2 in zip(statedict1, statedict2):
+            if p1.shape == p2.shape:
+                p1.copy_(p2)
+
+    def set_data(self,
+                 X_train: np.ndarray, y_train: np.ndarray,
+                 X_test: np.ndarray = None, y_test: np.ndarray = None) -> None:
+        """
+        Sets data for ensemble training (useful for iterative training)
+        """
+        if X_test is None or y_test is None:
+            X_train, X_test, y_train, y_test = train_test_split(
+                X_train, y_train, test_size=self.kdict.get("test_size", 0.15),
+                shuffle=True, random_state=0)
+        self.X_train = X_train
+        self.y_train = y_train
+        self.X_test = X_test
+        self.y_test = y_test
+
+    def run(self) -> Tuple[Type[torch.nn.Module], Dict, Dict]:
+        """
+        Trains a baseline model and ensemble of models
+        """
+        basemodel = self.train_baseline()
+        ensemble, ensemble_aver = self.train_ensemble(basemodel)
+        return basemodel, ensemble, ensemble_aver
+
+
+class ensemble_predictor:
+
+    """
+    Predicts mean and variance/uncertainty in image pixels
+    and (optionally) coordinates with ensemble of models
+
+    Args:
+        predictive_model (pytorch object):
+            model skeleton (can have randomly initialized weights)
+        ensemble (Dict):
+            nested dictionary with weights of each model in the ensemble
+        calculate_coordinates (bool):
+            computes atomic coordinates for each prediction
+        **eps (float): DBSCAN epsilon for clustering coordinates
+        **threshold (float):
+            value at which a neural network output is thresholded
+            for calculating coordinates
+        **num_classes (float): number of classes in the classification scheme
+        **downsample_factor (int): image downsampling (max_size / min_size) in NN
+    """
+
+    def __init__(self,
+                 predictive_model: Type[torch.nn.Module],
+                 ensemble: Dict[int, Dict[str, torch.Tensor]],
+                 calculate_coordinates: bool = False, **kwargs: Dict) -> None:
+
+        self.use_gpu = torch.cuda.is_available()
+
+        self.ensemble = ensemble
+        self.predictive_model = predictive_model
+
+        self.num_classes = kwargs.get("num_classes")
+        if self.num_classes is None:
+            hookF = [Hook(layer[1]) for layer in list(predictive_model._modules.items())]
+            mock_forward(predictive_model)
+            self.num_classes = [hook.output.shape for hook in hookF][-1][1]
+        self.downsample_factor = kwargs.get("downsample_factor")
+        if self.downsample_factor is None:
+            hookF = [Hook(layer[1]) for layer in list(predictive_model._modules.items())]
+            mock_forward(predictive_model)
+            imsize = [hook.output.shape[-1] for hook in hookF]
+            self.downsample_factor = max(imsize) / min(imsize)
+
+        self.calculate_coordinates = calculate_coordinates
+        if self.calculate_coordinates:
+            self.eps = kwargs.get("eps", 0.5)
+            self.thresh = kwargs.get("threshold", 0.5)
+
+    def predict(self,
+                x_new: np.ndarray
+                ) -> Tuple[Tuple[np.ndarray, np.ndarray],
+                           Union[Tuple[np.ndarray, np.ndarray], Tuple[None, None]]]:
+        """
+        Runs ensemble decoding for a single batch
+
+        Args:
+            x_new (numpy array): batch of images
+        """
+        x_new = img_pad(x_new, self.downsample_factor)
+        batch_dim, img_h, img_w = x_new.shape
+        nn_output_ensemble = np.zeros((
+            len(self.ensemble), batch_dim, img_h, img_w, self.num_classes))
+        for i, w in self.ensemble.items():
+            self.predictive_model.load_state_dict(w)
+            self.predictive_model.eval()
+            _, nn_output = predictor(
+                self.predictive_model,
+                nb_classes=self.num_classes,
+                downsampling=self.downsample_factor,
+                use_gpu=self.use_gpu,
+                verbose=False).decode(x_new, num_batches=1)
+            nn_output_ensemble[i] = nn_output
+        nn_output_mean = np.mean(nn_output_ensemble, axis=0)
+        nn_output_var = np.var(nn_output_ensemble, axis=0)
+        coord_mean, coord_var = None, None
+        if self.calculate_coordinates:
+            coord_mean, coord_var = ensemble_locate(
+                nn_output_ensemble, eps=self.eps, threshold=self.thresh)
+        return (nn_output_mean, nn_output_var), (coord_mean, coord_var)
+
+    def run(self,
+            imgdata: np.ndarray,
+            **kwargs: Dict
+            ) -> Tuple[Tuple[np.ndarray, np.ndarray],
+                       Union[Tuple[np.ndarray, np.ndarray], Tuple[None, None]]]:
+        """
+        Runs decoding with ensemble of models in a batch-by-batch fashion
+
+        Args:
+            imgdata (numpy array): 2D experimental image or 3D image stack
+            **num_batches (int): number of batches
+                (for large datasets to make sure everything fits into memory)
+        """
+        if np.ndim(imgdata) == 2:
+            imgdata = np.expand_dims(imgdata, axis=0)
+        imgdata = img_pad(imgdata, self.downsample_factor)
+        num_batches = kwargs.get("num_batches", 10)
+        batch_size = len(imgdata) // num_batches
+        if batch_size < 1:
+            batch_size = num_batches = 1
+        img_mu_all = np.zeros((*imgdata.shape[0:3], self.num_classes))
+        img_var_all = np.zeros(img_mu_all.shape)
+        coord_mu_all, coord_var_all = None, None
+        if self.calculate_coordinates:
+            coord_mu_all = np.zeros((imgdata.shape[0], 3))
+            coord_var_all = np.zeros(coord_mu_all.shape)
+
+        for i in range(num_batches):
+            print("\rBatch {}/{}".format(i+1, num_batches), end="")
+            x_i = imgdata[i*batch_size:(i+1)*batch_size]
+            (img_mu_i, img_var_i), (coord_mu_i, coord_var_i) = self.predict(x_i)
+            img_mu_all[i*batch_size:(i+1)*batch_size] = img_mu_i
+            img_var_all[i*batch_size:(i+1)*batch_size] = img_var_i
+            if self.calculate_coordinates:
+                coord_mu_all[i*batch_size:(i+1)*batch_size] = coord_mu_i
+                coord_var_all[i*batch_size:(i+1)*batch_size] = coord_var_i
+        x_i = imgdata[(i+1)*batch_size:]
+        if len(x_i) > 0:
+            (img_mu_i, img_var_i), (coord_mu_i, coord_var_i) = self.predict(x_i)
+            img_mu_all[(i+1)*batch_size:] = img_mu_i
+            img_var_all[(i+1)*batch_size:] = img_var_i
+            if self.calculate_coordinates:
+                coord_mu_all[(i+1)*batch_size:] = coord_mu_i
+                coord_var_all[(i+1)*batch_size:] = coord_var_i
+
+        return (img_mu_all, img_var_all), (coord_mu_all, coord_var_all)
+
+
+def ensemble_locate(nn_output_ensemble: np.ndarray,
+                    **kwargs: Dict) -> Tuple[np.ndarray, np.ndarray]:
+    """
+    Finds coordinates for each ensemble predictions
+    (basically, an atomnet.locator for ensembles)
+
+    Args:
+        nn_output_ensembles (numpy array):
+            5D numpy array with ensemble predictions
+        **eps (float):
+            DBSCAN epsilon for clustering coordinates
+        **threshold (float):
+            threshold value for atomnet.locator
+
+    Returns:
+        Mean and variance for every detected atom/defect/particle coordinate
+    """
+    eps = kwargs.get("eps", 0.5)
+    thresh = kwargs.get("threshold", 0.5)
+    coord_mean_all = {}
+    coord_var_all = {}
+    for i in range(nn_output_ensemble.shape[1]):
+        coordinates = {}
+        nn_output = nn_output_ensemble[:, i]
+        for i2, img in enumerate(nn_output):
+            coord = locator(thresh).run(img[None, ...])
+            coordinates[i2] = coord[0]
+        _, coord_mean, coord_var = cluster_coord(coordinates, eps)
+        coord_mean_all[i] = coord_mean
+        coord_var_all[i] = coord_var
+    return coord_mean_all, coord_var_all
